﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Globalization;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using WpfApplication1.Models;
using XnaMapGenerator3D.Services;
using Color = Microsoft.Xna.Framework.Color;

namespace XnaMapGenerator3D.Models
{
    public class MyModel
    {
        public Model Model { get; set; }
        public Vector3 Position { get; set; }
        
        public MyModel(Model model, Vector3 position)
        {
            Model = model;
            Position = position;
        }
    }

    public class Center : IEquatable<Center>, IMapItem
    {
        public int Index { get; set; }

        public int Key { get; set; }
        public Vector3 Point { get; set; }  // location
        public Boolean Water { get; set; }  // lake or ocean
        public Boolean Land { get { return !Water; } set { Water = !value; } }  // lake or ocean
        public Boolean Ocean { get; set; }  // ocean
        public Boolean ShallowWater { get; set; }
        public Boolean Coast { get; set; }  // land polygon touching an ocean
        public Boolean Border { get; set; }  // at the edge of the map
        public Vector3 Normal { get; set; }
        
        public Biome Biome { get; set; }
        Polygon Polygon { get; set; }
        
        public float Elevation
        {
            get { return this.Point.Y; }
            set { this.Point = new Vector3(Point.X, (float)value, Point.Z); }
        }
        public double Moisture { get; set; }  // 0.0-1.0

        List<MyModel> trees;
        private Texture2D treeTexture;
        private readonly BGame _game;

        public HashSet<Center> Neighbours { get; set; }
        public HashSet<Edge> Borders { get; set; }
        public HashSet<Corner> Corners { get; set; }

        public Center(BGame game, float x , float y, float z)
        {
            //polyNorm = new Vector3D(0, 0, 0);
            Point = new Vector3(x, y , z);
            Key = Point.GetHashCode();
            Index = Point.GetHashCode();
            trees = new List<MyModel>();
            treeTexture = game.Content.Load<Texture2D>("texture2");

            Water = Coast = Ocean = Border = false;
            Moisture = 0.0f;

            Neighbours = new HashSet<Center>(new CenterComparer());
            Borders = new HashSet<Edge>(new EdgeComparer());
            Corners = new HashSet<Corner>(new CornerComparer());
            
            _game = game;
        }

        #region Methods
        public bool Equals(Center other)
        {
            return this.Point.X == other.Point.X && this.Point.Y == other.Point.Y && this.Point.Z == other.Point.Z;
        }

        public void OrderCorners()
        {
            var currentCorner = Corners.First();
            var ordered = new List<Corner>(Corners.Count);
            Edge ed;

            ordered.Add(currentCorner);
            do
            {
                ed = currentCorner.Protrudes.FirstOrDefault(x => Borders.Contains(x) && !(ordered.Contains(x.VoronoiStart) && (ordered.Contains(x.VoronoiEnd))));

                if (ed != null )
                {
                    Corner newdot = ed.Corners.FirstOrDefault(x => !ordered.Contains(x) && x.Touches.Any(a => a == this));
                    if (newdot != null)
                    {
                        ordered.Add(newdot);
                        currentCorner = newdot;
                    }
                    else
                    {
                        ed = null;
                    }
                }
            } while (ed != null);

            if(ordered.Count != Corners.Count)
            {
                
            }

            Corners.Clear();

            foreach (var corner in ordered)
            {
                Corners.Add(corner);
            }
        }

        public void FixBorders(MapGenService mapGen)
        {
            var ms = from p in Borders.SelectMany(x => x.Corners)
                     group p by p.Point
                         into grouped
                         select new { Point = grouped.Key, count = grouped.Count() };

            var fpoint = ms.FirstOrDefault(x => x.count == 1);
            var spoint = ms.LastOrDefault(x => x.count == 1);

            if (fpoint != null & spoint != null)
            {
                Corner p1 = Corners.FirstOrDefault(x => x.Point == fpoint.Point);
                Corner p2 = Corners.FirstOrDefault(x => x.Point == spoint.Point);

                if (p1 == null || p2 == null)
                    return;

                IFactory fact = new DataFactory(mapGen);
                Edge e = fact.EdgeFactory(
                    p1,
                    p2,
                    this, null);

                e.MapEdge = true;

                p1.Protrudes.Add(e);
                p2.Protrudes.Add(e);

                this.Border = this.Ocean = this.Water = true;
                e.VoronoiStart.Border = e.VoronoiEnd.Border = true;
                e.VoronoiStart.Elevation = e.VoronoiEnd.Elevation = 0.0f;

                this.Borders.Add(e);
            }
        }
        
        public void SetBiome()
        {
            Biome = BiomeTypes.BiomeSelector(Ocean, Water, Coast, Elevation, Moisture);
        }

        private void AddModel(string name)
        {
            if (!BGame.AddVegetation)
                return;

            Model newTree = LoadModel(name);
            trees.Add(new MyModel(newTree,this.Point));

            foreach (Corner co in this.Corners)
            {
                trees.Add(new MyModel(newTree, Vector3.Lerp(this.Point, co.Point, 0.4f)));
                trees.Add(new MyModel(newTree, Vector3.Lerp(this.Point, co.Point, 0.8f)));
            }

            foreach (Edge bo in this.Borders)
            {
                trees.Add(new MyModel(newTree, Vector3.Lerp(this.Point, new Vector3(bo.Midpoint.X, this.Point.Y, bo.Midpoint.Z), 0.4f)));
                trees.Add(new MyModel(newTree, Vector3.Lerp(this.Point, new Vector3(bo.Midpoint.X, this.Point.Y, bo.Midpoint.Z), 0.8f)));
            }
        }

        private Model LoadModel(string assetName)
        {

            Model newModel = _game.Content.Load<Model>(assetName); foreach (ModelMesh mesh in newModel.Meshes)
                foreach (ModelMeshPart meshPart in mesh.MeshParts)
                    meshPart.Effect = _game.ShaderEffect.Clone();
            return newModel;
        }

        public void CalculateNormals()
        {
            var sx = Corners.Sum(x => x.Normal.X) / Corners.Count;
            var sy = Corners.Sum(x => x.Normal.Y) / Corners.Count;
            var sz = Corners.Sum(x => x.Normal.Z) / Corners.Count;

            Normal = new Vector3(sx, sy, sz);
        }

        public void GetPolygons()
        {
            Polygon = new Polygon(this.Corners, this);
        }
        #endregion
        
        public void Draw(BGame game, Matrix viewMatrix, Matrix projectionMatrix)
        {
            game.ShaderEffect.TextureEnabled = false;
            //game.ShaderEffect.Texture = _texture;
            game.ShaderEffect.VertexColorEnabled = true;

            foreach (var pass in game.ShaderEffect.CurrentTechnique.Passes)
            {
                pass.Apply();
            }

            Polygon.Draw(game.GraphicsDevice);

            if (trees.Count > 0)
            {
                foreach (MyModel mm in trees)
                {
                    Matrix worldMatrix = Matrix.CreateScale(0.5f, 0.5f, 0.5f) * Matrix.CreateRotationY(MathHelper.Pi) *
                                         Matrix.CreateTranslation(mm.Position);

                    Matrix[] transforms = new Matrix[mm.Model.Bones.Count];
                    mm.Model.CopyAbsoluteBoneTransformsTo(transforms);

                    foreach (ModelMesh mesh in mm.Model.Meshes)
                    {
                        foreach (BasicEffect currentEffect in mesh.Effects)
                        {
                            //_game.GraphicsDevice.DepthStencilState = mm.DepthState;
                            currentEffect.Texture = treeTexture;
                            currentEffect.TextureEnabled = true;
                            currentEffect.World = transforms[mesh.ParentBone.Index]*worldMatrix;
                            currentEffect.View = viewMatrix;
                            currentEffect.Projection = projectionMatrix;
                        }
                        mesh.Draw();
                    }
                }
            }
        }

        
    }
}
